#include <stdio.h>
#include <v_lib.h>
#include "wmf_ops.h"

/*	Placeable metafile signature */
#define	PMPHKEY		0x9AC6CDD7

#pragma pack(push)
#pragma pack(2)
/*	Placeable metafile header */
typedef struct pmfh_s
{
	u32_t	key;
	u16_t	emptywmf;
	u16_t	left, top, right, bottom;
	u16_t	inch;
	u32_t	reserved;
	u16_t	checksum;
}	pmfh_t;

/*	Metafile header */
typedef struct 	metaheader_s
{
    u16_t       type;
    u16_t       header_size;
    u16_t       version;
    u32_t      	size;
    u16_t       num_objects;
    u32_t       max_record;
    u16_t       no_parameters;
}	metaheader_t, *metaheader_p;

#pragma pack(pop)

char_p	V_LOCAL
wmfRecordName(u16_t code)
{
	int i = 0;
	while(metafunc[i].code)
		if(metafunc[i++].code==code)
			break;
	return metafunc[i].name;
}

void	V_LOCAL
wmfDump(char *name)
{
	file_t f;
	pmfh_t aph;
	u32_t rec_size;
	u16_t rec_func;
	u16_t *arg;
	metaheader_t mh;
	if(!name)
		return;
	f=fileOpenRead(name);
	if(!f)
		return;
	aph.key=fileGetLittleLong(f);
	aph.emptywmf=fileGetLittleShort(f);
	aph.left=fileGetLittleShort(f);
	aph.top=fileGetLittleShort(f);
	aph.right=fileGetLittleShort(f);
	aph.bottom=fileGetLittleShort(f);
	aph.inch=fileGetLittleShort(f);
	aph.reserved=fileGetLittleLong(f);
	aph.checksum=fileGetLittleShort(f);
	if(aph.key==PMPHKEY)
	{
		printf("File '%s' is Adobe (placeable) metafile.\n"
			"APMHeader contents:\n"
			"Key                %X\n"
			"Handle             %u\n"
			"RECT               {%d,%d,%d,%d} (%d X %d)\n"
			"Inch               %d\n"
			"Checksum           %u\n",
			name, aph.key, aph.emptywmf, aph.left, aph.top, aph.right,
			aph.bottom, aph.right-aph.left, aph.bottom-aph.top,
			aph.inch, aph.checksum );
	}
	else
	{
		fileSeek(f,0,SEEK_CUR);
		printf("File '%s' is Windows 3.0 / 3.1 metafile.\n",name);
	}
    mh.type=fileGetLittleShort(f);
    mh.header_size=fileGetLittleShort(f);
    mh.version=fileGetLittleShort(f);
    mh.size=fileGetLittleLong(f);
    mh.num_objects=fileGetLittleShort(f);
    mh.max_record=fileGetLittleLong(f);
    mh.no_parameters=fileGetLittleShort(f);
	if(mh.type!=1 || mh.header_size!=9)
	{
		conPrintf("File '%s' damaged: mh.type!=1 || mh.header_size!=9\n",
			name);
		fileClose(f);
		return;
	}
	printf("Header contents:\n"
		"Type               0x%0004X\n"
		"HeaderSize         %u\n"
		"Version            0x%0004X\n"
		"Size               %lu\n"
		"NoObjects          %u\n"
		"MaxRecord          %lu\n"
		"NoParameters       %u\n",
		mh.type,mh.header_size,mh.version,mh.size,mh.num_objects,
		mh.max_record, mh.no_parameters);
	printf("Function list:\n");
	arg=(u16_t*)malloc(sizeof(u16_t)*(mh.max_record+1));
	for(;;)
	{
		rec_size=fileGetLittleLong(f);
		rec_func=fileGetLittleShort(f);
		if(rec_size==3 && rec_func==0)
			break;
		printf("Code 0x%0004X, function '%s' ",rec_func,
			wmfRecordName(rec_func));
		rec_size-=3;
		if(rec_size>0)
		{
			int warg;
			fileRead(f,arg,2*rec_size);
			printf("(");
			for(warg=rec_size-1;warg>=0;warg--)
				printf("%d ",arg[warg]);
			puts(")");
		}
		else
			puts("(no arguments)");
	}
	free(arg);
	fileClose(f);
}

int main(int argc, char *argv[])
{
	if(argc!=2)
		return -1;
	logStartConsoleLog();
	wmfDump(argv[1]);
	return 0;
}
